package arbol.avl.modelo;

import arbol.binariodebusqueda.modelo.Comparable;
import arbol.binariodebusqueda.modelo.Logical;
import arbol.binariodebusqueda.modelo.Nodo;

public class ArbolAVL {
	private Nodo raiz;

	public ArbolAVL() {
		raiz=null;
	}
	
	public Nodo getRaiz(){
		return raiz;
	}
	
	/**
	 * ROTACIONES
	 */
	/**
	 * Rotacion simple: Izquierda-Izquierda
	 */
	private Nodo rotacionII(Nodo n, Nodo n1){
		n.setIzquierdo(n1.getDerecho());
		n1.setDerecho(n);
		//Actualización de los factores de equilibrio
		if (n1.getFe()==-1)//Se cumple en la insercion
		{
			n.setFe(0);
			n1.setFe(0);
		}else{
			n.setFe(-1);
			n1.setFe(1);
		}
		return n1;
	}
	/**
	 * Rotacion simple: Derecha-Derecha
	 * @param n
	 * @param n1
	 * @return
	 */
	private Nodo rotacionDD(Nodo n, Nodo n1){
		n.setDerecho(n1.getIzquierdo());
		n1.setIzquierdo(n);
		//Actualización de los factores de equilibrio
		if(n1.getFe() == +1)//Se cumple en la inserción
		{
			n.setFe(0);
			n1.setFe(0);
		}else{
			n.setFe(+1);
			n1.setFe(-1);			
		}
		return n1;
	}
	/**
	 * Rotacion doble: Izquierda-Derecha
	 * @param n
	 * @param n1
	 * @return
	 */
	
	private Nodo rotacionID(Nodo n, Nodo n1){
		Nodo n2;
		n2 = (Nodo)n1.getDerecho();
		n.setIzquierdo(n2.getDerecho());
		n2.setDerecho(n);
		n1.setDerecho(n2.getIzquierdo());
		n2.setIzquierdo(n1);
		//Actualización de los factores de equilibrio
		if(n2.getFe()==+1)
			n1.setFe(-1);
		else
			n1.setFe(0);
		if(n2.getFe()==-1)
			n.setFe(1);
		else
			n.setFe(0);
		n2.setFe(0);
		return n2;
	}
	/**
	 * Rotacion doble: Derecha-Izquierda
	 * @param n
	 * @param n1
	 * @return
	 */
	private Nodo rotacionDI(Nodo n, Nodo n1){
		Nodo n2;
		n2 = (Nodo)n1.getIzquierdo();
		n.setDerecho(n2.getIzquierdo());
		n2.setIzquierdo(n);
		n1.setIzquierdo(n2.getDerecho());
		n2.setDerecho(n1);
		//Actualización de los factores de equilibrio
		if(n2.getFe()==+1)
			n.setFe(-1);
		else
			n.setFe(0);
		if(n2.getFe()==-1)
			n1.setFe(1);
		else
			n1.setFe(0);
		n2.setFe(0);
		return n2;
	}
	/**
	 * Inserta un elemento en el arbol avl, llama al metodo recursivo insertarAvl
	 * @param valor
	 * @throws Exception
	 */
	public void insertar(Object valor) throws Exception{
		Comparable dato;
		Logical h = new Logical(false);//intercambio un valor booleano
		dato = (Comparable)valor;
		raiz = insertarAvl(raiz, dato, h);
	}
	/**
	 * Metodo recursivo
	 * @param raiz, raiz
	 * @param dt dato
	 * @param h
	 * @return Nodo
	 * @throws Exception
	 */
	private Nodo insertarAvl(Nodo raiz, Comparable dt, Logical h) throws Exception{
		Nodo n1;
		if (raiz == null){
			raiz = new Nodo(dt);
			h.setLogical(true);
			
		}
		else if (dt.esMenor(raiz.getValor())){
			Nodo iz;
			iz = insertarAvl((Nodo) raiz.getIzquierdo(), dt, h);
			raiz.setIzquierdo(iz);
			//regreso por los nodos del camino de búsqueda
			if(h.booleanValue()){
				//decrementa el fe por aumentar la altura de rama izquireda
				switch(raiz.getFe()){
				case 1:
					raiz.setFe(0);
					h.setLogical(false);
					break;
				case 0:
					raiz.setFe(-1);
					break;
				case -1://aplicar rotación a la izquierda
					n1 = (Nodo)raiz.getIzquierdo();
					if(n1.getFe()==-1)
						raiz = rotacionII(raiz, n1);
					else
						raiz = rotacionID(raiz, n1);
					h.setLogical(false);
				}
			}
		}
		else if (dt.esMayor(raiz.getValor())){
			Nodo dr;
			dr = insertarAvl((Nodo) raiz.getDerecho(), dt, h);
			raiz.setDerecho(dr);
			//regreso por los nodos del camino de búsqueda
			if(h.booleanValue()){
				//incrementa el fe por aumentar la altura de rama izquireda
				switch(raiz.getFe()){
				case 1://aplicar rotacion a la derecha
					n1 = (Nodo) raiz.getDerecho();
					if(n1.getFe() == +1)
						raiz = rotacionDD(raiz, n1);
					else
						raiz = rotacionDI(raiz, n1);
					h.setLogical(false);
					break;
				case 0:
					raiz.setFe(+1);
					break;
				case -1:
					raiz.setFe(0);
					h.setLogical(false);
				}
			}
			
		}
		else
			throw new Exception("No pueden haber claves repetidas");
		return raiz;
	}
	/**
	 * Elimina un nodo del arbol avl
	 * @param valor valor a eliminar
	 * @throws Exception
	 */
	
	public void eliminar (Object valor) throws Exception{
		Comparable dato;
		dato = (Comparable) valor;
		Logical flag = new Logical(false);
		raiz = borrarAvl(raiz, dato, flag);
	}
	/**
	 * Metodo recursivo para eliminar un nodo
	 * @param r nodo raiz
	 * @param clave elemento clave a borrar
	 * @param cambiaAltura tipo Logical
	 * @return
	 * @throws Exception
	 */
	private Nodo borrarAvl(Nodo r, Comparable clave, Logical cambiaAltura) throws Exception{
		if (r==null){
			throw new Exception("Nodo no encontrado");
		}
		else if(clave.esMenor(r.getValor())){
			Nodo iz;
			iz = borrarAvl((Nodo) r.getIzquierdo(), clave, cambiaAltura);
			r.setIzquierdo(iz);
			if(cambiaAltura.booleanValue())
				r = equilibrar1(r, cambiaAltura);
		}
		else if(clave.esMayor(r.getValor())){
			Nodo dr;
			dr = borrarAvl((Nodo) r.getDerecho(), clave, cambiaAltura);
			r.setDerecho(dr);
			if (cambiaAltura.booleanValue())
				r = equilibrar2(r, cambiaAltura);
		}
		else//Nodo encontrado
		{
			Nodo q;
			q = r;//Nodo a quitar del arbol
			if(q.getIzquierdo() == null){
				r = (Nodo) q.getDerecho();
				cambiaAltura.setLogical(true);
			}
			else if (q.getDerecho() == null){
				r = (Nodo) q.getIzquierdo();
				cambiaAltura.setLogical(true);
			}
			else{
				//tiene rama izquierda y derecha
				Nodo iz;
				iz = reemplazar(r, (Nodo) r.getIzquierdo(), cambiaAltura);
				r.setIzquierdo(iz);
				if (cambiaAltura.booleanValue())
					r = equilibrar1(r, cambiaAltura);
			}
			
			q = null;			
		}
		return r;
	}
	/**
	 * Reemplaza el nodo a eliminar por el nodo más a la izquierda del hijo derecho
	 * @param n
	 * @param act
	 * @param cambiaAltura
	 * @return
	 */
	private Nodo reemplazar(Nodo n, Nodo act, Logical cambiaAltura){
		if(act.getDerecho() == null){
			Nodo d;
			d = reemplazar(n, (Nodo)act.getDerecho(), cambiaAltura);
			act.setDerecho(d);
			if (cambiaAltura.booleanValue())
				act = equilibrar2(act, cambiaAltura);
		}else{
			n.setValor(act.getValor());
			n = act;
			act = (Nodo) act.getIzquierdo();
			n = null;
			cambiaAltura.setLogical(true);
		}
		return act;
	}
	
	private Nodo equilibrar1(Nodo n, Logical cambiaAltura){
		Nodo n1;
		switch(n.getFe()){
		case -1:
			n.setFe(0);
			break;
		case 0:
			n.setFe(1);
			cambiaAltura.setLogical(false);
			break;
		case +1:
			//Se aplica un tipo de rotación derecha
			n1 = (Nodo) n.getDerecho();
			if(n1.getFe()>=0){
				if(n1.getFe()==0)//La altura no vuelve a disminuir
					cambiaAltura.setLogical(false);
				n = rotacionDD(n, n1);
			}
			else
				n = rotacionDI(n, n1);
			break;
		}
		return n;
	}
	
	private Nodo equilibrar2(Nodo n, Logical cambiaAltura){
		Nodo n1;
		switch(n.getFe()){
		case -1:
			//Se aplica un tipo de rotacion izquierda
			n1 = (Nodo) n.getIzquierdo();
			if(n1.getFe()<=0){
				if (n1.getFe()==0)
					cambiaAltura.setLogical(false);
				n = rotacionDD(n, n1);
			}
			else
				n = rotacionID(n, n1);
			break;
		case 0:
			n.setFe(-1);
			cambiaAltura.setLogical(false);
			break;
		case +1:
			n.setFe(0);
			break;
		}
		return n;
	}
	
	public void preorden() {
		preorden(raiz);
	}
	private void preorden(Nodo aux) {
		if (aux != null) {
			visitar(aux);
			preorden(aux.getIzquierdo());
			preorden(aux.getDerecho());
		}
	}

	private void visitar(Nodo aux) {
		System.out.print(aux.getValor() + " ");
	}


}
