package arbol.binario.modelo;

public class ArbolBinario {
	/**
	 * Raiz del arbol binario
	 */
	private Nodo raiz;

	/**
	 * Constructor por defecto
	 */
	public ArbolBinario() {
		raiz = null;
	}

	/**
	 * Inserta un nodo en el arbol binario
	 * 
	 * @param valor
	 *            valor a insertar
	 */
	public void inicializar() {
		Nodo nodoA = new Nodo("A");
		Nodo nodoB = new Nodo("B");
		Nodo nodoC = new Nodo("C");
		Nodo nodoD = new Nodo("D");
		Nodo nodoE = new Nodo("E");
		Nodo nodoF = new Nodo("F");
		Nodo nodoG = new Nodo("G");

		raiz = nodoA;
		raiz.setIzquierdo(nodoB);
		raiz.setDerecho(nodoC);
		/**
		 * @param args
		 */
		nodoB.setIzquierdo(nodoD);
		nodoB.setDerecho(nodoE);

		nodoC.setIzquierdo(nodoF);
		nodoC.setDerecho(nodoG);

	}

	/**
	 * Invoca al método recursivo
	 */
	public void preorden() {
		preorden(raiz);
	}

	/**
	 * Método recursivo de recorrido en pre-orden
	 * 
	 * @param aux
	 *            referencia a un nodo
	 */
	private void preorden(Nodo aux) {
		if (aux != null) {
			visitar(aux);
			preorden(aux.getIzquierdo());
			preorden(aux.getDerecho());
		}
	}

	private void visitar(Nodo aux) {
		System.out.print(aux.getValor() + " ");
	}

	/**
	 * Invoca al método recursivo
	 */
	public void inorden() {
		inorden(raiz);
	}

	/**
	 * Método recursivo de entre-orden
	 * 
	 * @param aux
	 *            referencia a un nodo
	 */
	private void inorden(Nodo aux) {
		if (aux != null) {
			inorden(aux.getIzquierdo());
			visitar(aux);
			inorden(aux.getDerecho());
		}
	}

	/**
	 * Invoca al método recursivo
	 */
	public void postorden() {
		postorden(raiz);
	}

	/**
	 * Método recursivo post-orden
	 * 
	 * @param aux
	 *            referencia a un nodo
	 */
	private void postorden(Nodo aux) {
		if (aux != null) {
			postorden(aux.getIzquierdo());
			postorden(aux.getDerecho());
			visitar(aux);
		}
	}

	public Nodo getRaiz() {
		return raiz;
	}

	/**
	 * Recorrido por niveles
	 */
	public void recorridoPorNiveles() {
		utilidades.Cola<Nodo> cola = new utilidades.Cola<Nodo>();
		cola.encolar(raiz);
		while (!cola.esVacia()) {
			Nodo aux = cola.frente();
			visitar(aux);
			if (aux.getIzquierdo() != null)
				cola.encolar(aux.getIzquierdo());
			if (aux.getDerecho() != null)
				cola.encolar(aux.getDerecho());
			cola.desencolar();
		}
	}

	/**
	 * Recorrido preorden version iterativa, utiliza una pila
	 */
	public void preordenIterativo() {
		utilidades.Pila<Nodo> pila = new utilidades.Pila<Nodo>();
		pila.apilar(raiz);
		while (!pila.esVacia()) {
			Nodo aux = pila.cima();
			visitar(aux);
			pila.retirar();
			if (aux.getDerecho() != null)
				pila.apilar(aux.getDerecho());
			if (aux.getIzquierdo() != null)
				pila.apilar(aux.getIzquierdo());
	
		}
	}

	/**
	 * Recorrido inorden version iterativa, utiliza una pila
	 * 
	 */
	public void inordenIterativo() {
		utilidades.Pila<Nodo> pila = new utilidades.Pila<Nodo>();
		pila.apilar(raiz);
		Nodo aux = raiz.getIzquierdo();
		while (aux != null || !pila.esVacia()) {
			if (aux != null) {
				pila.apilar(aux);
				aux = aux.getIzquierdo();
			} else {
				aux = pila.cima();
				pila.retirar();
				visitar(aux);
				aux = aux.getDerecho();
			}
		}
	}

	/**
	 * Recorrido postorden version iterativa, utiliza una pila.
	 */
	public void postOrdenIterativo() {
		utilidades.Pila<Nodo> pila = new utilidades.Pila<Nodo>();
		Nodo aux = raiz;
		Nodo q = raiz;
		while (aux != null) {
			// avanza por izquierda y apila los nodos
			while (aux.getIzquierdo() != null) {
				pila.apilar(aux);
				aux = aux.getIzquierdo();
			}
			while (aux != null
					&& (aux.getDerecho() == null || aux.getDerecho() == q)) {
				visitar(aux);
				q = aux;
				if (pila.esVacia())
					return;
				aux = pila.cima();
				pila.retirar();
	
			}
			pila.apilar(aux);
			aux = aux.getDerecho();
		}
	}
}
