package arbol.graficar;

import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

import javax.swing.*;
import javax.swing.border.LineBorder;

import arbol.binariodebusqueda.modelo.Comparable;
import arbol.binariodebusqueda.modelo.Nodo;

/**
 * Representa el tablero sobre el cual se dibujar el arbol. Este tablero 
 * puede ser agregado desde cualquier JFrame. La logica es la siguiente:
 * 1. El metodo dibujar(raiz), recibe la raiz del arbol binario que se debe dibujar
 * 2. Se hace el recorrido en preorden del arbol
 * 3. En lugar de visitar cada nodo, se lo inserta en un arbol binario interno auxiliar
 * 4. Conforme se inserta cada nodo, se va calculando las coordenadas x,y del nodo y de su nodo padre
 * 5. Se agrega cada nodo, en un vector de vertices
 * 6. Finalmente, el metodo dibujarArbol() es quien dibuja el arbol, recorriendo cada vertice del vector
 * 
 * @author libardo
 * 
 */
public class Tablero extends JComponent {
	/**
	 * Largo y ancho de un nodo en pixeles
	 */
	private static int LARGO_NODO = 25;
	/**
	 * Lista de vertices a ser dibujados
	 */
	private ArrayList<Vertice> vertices;
	/**
	 * Raiz del arbol binario
	 *//**
	 * Representa la pila como tal con sus operaciones principales de apilar y
	 * retirar
	 * 
	 * @author Libardo
	 * 
	 * @param <T>
	 *            tipo de objeto a almacenar en la pila
	 */
	private Nodo raiz;

	public Tablero() {
		raiz = null;
		vertices = new ArrayList<Vertice>();
	}

	/**
	 * Metodo publico, el unico, que permite recorrer el arbol binario en
	 * preorden, y cada nodo agregarlo al vector de nodos, para finalmente
	 * graficarlos.
	 * 
	 * @param raiz
	 */
	public void dibujar(Nodo raiz) {
		preorden(raiz);
	}

	/**
	 * Metodo que nos permite pintar el componente.
	 */
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;

		// Color de fondo blanco
		g2.setColor(Color.WHITE);
		g2.fillRect(0, 0, this.getWidth(), this.getHeight());
		// Color de la fuente y del borde externo
		g2.setColor(Color.BLACK);
		g2.drawRect(0, 0, this.getWidth(), this.getHeight());
		dibujarArbol(g2);
		g2.dispose();

	}

	/**
	 * Este metodo pinta el arbol binario en el componente utilizando el objeto
	 * Graphics2D pasado como parametro. La grafica utiliza los valores que se
	 * encuentran en el vector de este objeto.
	 * 
	 * @param g2
	 *            Objeto sobre el cual se dibuja
	 */

	private void dibujarArbol(Graphics2D g2) {

		g2.setFont(new Font("_Sans", Font.BOLD, 10));
		g2.drawString("Autor: W. Libardo Pantoja Y.",10,10);
		g2.setColor(Color.BLACK);

		int x1 = 0, y1 = 0, x2 = 0, y2 = 0;
		// Recorrer cada nodo y pintarlo
		for (int i = 0; i < vertices.size(); i++) {
			if (vertices.get(i).getPuntoPadre() == null) {
				// RAIZ: no se dibuja linea
				// Dibujar circulo
				x1 = vertices.get(i).getPuntoNodo().getX();
				y1 = vertices.get(i).getPuntoNodo().getY();

				g2.drawOval(x1 - LARGO_NODO / 2, y1 - LARGO_NODO, LARGO_NODO,
						LARGO_NODO);
				//Pintar el texto del nodo
				g2.drawString(vertices.get(i).getDato(), x1 - LARGO_NODO / 2
						+ (int) (LARGO_NODO * 0.1), y1 - LARGO_NODO / 2);

			} else {
				// NO ES LA RAIZ
				
				//Dibujar la linea
				x1 = vertices.get(i).getPuntoPadre().getX();
				y1 = vertices.get(i).getPuntoPadre().getY();

				x2 = vertices.get(i).getPuntoNodo().getX();
				y2 = vertices.get(i).getPuntoNodo().getY();
				/**
				 * Representa la pila como tal con sus operaciones principales de apilar y
				 * retirar
				 * 
				 * @author Libardo
				 * 
				 * @param <T>
				 *            tipo de objeto a almacenar en la pila
				 */
				g2.drawLine(x1, y1, x2, y2);
				//Dibujar el circulo
				g2.drawOval(x2 - LARGO_NODO / 2, y2, LARGO_NODO, LARGO_NODO);
				//Pintar el texto del nodo
				g2.drawString(vertices.get(i).getDato(), x2 - LARGO_NODO / 2
						+ (int) (LARGO_NODO * 0.1), y2 + LARGO_NODO / 2);
			}
		}
	}

	/**
	 * Inserta un valor al arbol binario interno, y además, lo agrega al vector de
	 * vertices. Es similar al algoritmo de insertar de un arbol binario
	 * con la diferencia que al final agrega el nodo al vector de vertices
	 * que se deberán pintar
	 * 
	 * @param valor valor del nodo
	 */
	private void insertar(Object valor) {
		int factorX = 430;
		int factorY = 60;

		Comparable dato = (Comparable) valor;

		Nodo nuevo = new Nodo();
		nuevo.setValor(dato);
		int nivel = 1;
		int x = this.getWidth() / 2;
		Punto puntoPadre = null;
		if (raiz == null) {
			raiz = nuevo;
			Punto puntoNodo = new Punto(x, nivel * factorY + LARGO_NODO);
			agregarVertice(dato.toString(), puntoNodo, null);

		} else {
			Nodo anterior = null;
			Nodo aux = raiz;

			while (aux != null) {
				puntoPadre = new Punto(x, nivel * factorY + LARGO_NODO);

				nivel++;
				factorX = (int) (factorX * 0.53);
				this.setBackground(Color.blue);
				anterior = aux;
				if (dato.esMenor(aux.getValor())) {
					aux = aux.getIzquierdo();
					x = x - factorX;
				} else {
					aux = aux.getDerecho();
					x = x + factorX;
				}

			}
			if (dato.esMenor(anterior.getValor())) {
				anterior.setIzquierdo(nuevo);
			} else {
				anterior.setDerecho(nuevo);
			}
			Punto puntoNodo = new Punto(x, nivel * factorY);

			agregarVertice(dato.toString(), puntoNodo, puntoPadre);
		}
	}

	/**
	 * Agrega un nuevo vertice al vector de vertices
	 * 
	 * @param dato texto del nodo
	 * @param puntoNodo punto x,y del nodo 
	 * @param puntoPadre punto x,y del nodo padre
	 */
	private void agregarVertice(String dato, Punto puntoNodo, Punto puntoPadre) {
		Vertice v = new Vertice(dato, puntoNodo, puntoPadre);
		vertices.add(v);
	}

	/**
	 * Recorrido recursivo en preorden, cada nodo es insertado al 
	 * arbol binario
	 * 
	 * @param aux nodo de un subarbol
	 */
	private void preorden(Nodo aux) {
		if (aux != null) {
			insertar(aux.getValor());
			preorden(aux.getIzquierdo());
			preorden(aux.getDerecho());
		}
	}
}